#include "Game.h"
#include "global.h"

#include <strstream>
#include <stdio.h>
#include <stdlib.h>

bool Game::BeginPlay(char * Server, int Port)
{
	mIO->connect(Server, Port);
	mIO << "Player" << endl;
	{
		int X, Y;

		mIO >> X >> Y;
		mMapDimensions.SetX(X);
		mMapDimensions.SetY(Y);
	
		for (int i=1; i <= Y; i++) 
		{
			char MapBuffer[X];
    		mIO >> MapBuffer;
			for(int j = 1; j <= X; j++)
			{
				TileType val = (TileType)MapBuffer[j - 1];
				mMap.Set(j, i, val);
			}
		}
		ID id; Weight weight; Money money;
		mIO >> id >> weight >> money;
		mRoboto = mRobots.Add(new Robot (id, Point (), weight, money));
		mIO.read(&id, 1);

	}
	mMap.DumpArray();
	InnerLoop();
	return true;
}

void Game::InnerLoop() 
{
	int Offset = 0;
	static char Buffer[1024];
	//int count = 0;

	while (!mIO.eof())
	{

		// Handle Players
		do 
		{
			mIO.getline(Buffer + Offset, 1024 - Offset);
			Offset = HandleMessages (Buffer + Offset, !mIO);
		}
		while (!mIO && !mIO.eof());

		Offset = Buffer[0] = 0;

		// Handle PackageList
		do 
		{
			mIO.getline(Buffer + Offset, 1024 - Offset);
			Offset = HandleMessages (Buffer + Offset, !mIO);
		}
		while (!mIO && !mIO.eof());

		mRoboto->MakeMove(this);
	}
}

inline Robot* Game::UpdateRobot (ID RobotID, MoveType action, Package* package)
{
	Robot* robot = mRobots.Find(RobotID);
	robot->Update (action, package);
	return robot;
}

inline Package* Game::UpdatePackage(ID pkgid, MoveType action, Robot* robot)
{
	Package* package = mPackageList.Find(pkgid);
	return UpdatePackage (package, action, robot);
}

inline Package* Game::UpdatePackage(Package * package, MoveType action, Robot* robot)
{
	if(action == DROP)
	{
		if(robot->GetLocation() == package->GetDestination())
		//package has been delivered, remove from list
		{
			mPackageList.Remove(package);
		}
		else
		{
			package->SetLocation(robot->GetLocation());
			package->SetOwner(NULL);
		}
	}
	else
		package->SetOwner(robot);
	
	return package;
}

inline Package* Game::UpdatePackage(ID pkgid, Point destination, Weight weight, Point location)
{
	Package* package = mPackageList.Find(pkgid);
	if(package == NULL)
		package = mPackageList.Add(new Package(pkgid, location, weight, destination));
	else
		package->SetLocation(location);

	return package;
}

inline Robot * Game::UpdateRobot (ID RobotID, Point location)
{
	Robot* robot = mRobots.Find(RobotID);
	if(robot == NULL) 
		return mRobots.Add(new Robot(RobotID, location));
	else
		robot->SetLocation(location);

	return robot;
}

inline Robot * Game::UpdateRobot (ID RobotID, MoveType direction)
{
	Robot* robot = mRobots.Find(RobotID);
	robot->Update(direction);
	return robot;
}

int Game::HandlePlayerUpdateMessage(char * UpdateString, int nc)
{
	vector<string> tokens;
	string MsgStr(UpdateString);
	Tokenize(MsgStr, tokens, "#");
	for(unsigned int i = 0; i < tokens.size() - nc; i++)
	{
		istrstream TokenReader(tokens[i].c_str());
		ID RobotID;
		TokenReader >> RobotID;
		while(!TokenReader.eof())
		{
			MoveType action = NOOP;
			int x, y;
			ID pkgid = 0;
			Robot* robot = NULL;
			Package* package = NULL;
			TokenReader >> (char)action;
			switch(action)
			{
			case PICK:
			case DROP:
				TokenReader >> pkgid;
				package = mPackageList.Find (pkgid);
				robot = UpdateRobot (RobotID, action, package);
				UpdatePackage(package, action, robot);
				break;
			case NORTH:
			case SOUTH:
			case WEST:
			case EAST:
				UpdateRobot (RobotID, action);
				break;
			case POS:
				TokenReader >> x;
				TokenReader >> (char)action;
				TokenReader >> y;
				UpdateRobot (RobotID, Point (x, y));
				break;
			default:
				break;
			}
		}
			
	}
	
	strcpy(UpdateString, tokens[tokens.size() - 1].c_str());;
	return strlen(UpdateString);
}

int Game::HandlePackageUpdateMessage(char* UpdateString)
{
	string MsgStr(UpdateString);
	vector<string> tokens;

	Tokenize(MsgStr, tokens);
	unsigned int LeftOverTokens = tokens.size() % 4;
	unsigned int i = 0;

	while (i < (tokens.size() - LeftOverTokens))
	{
		ID pkgid; int x = 0, y = 0; Weight weight;

		//FIXME
		string NastyString;
		for (unsigned int j = i; j < i + 4; j++)
		{
			NastyString += tokens[j];
			NastyString += " ";
		}

		istrstream TokenReader(NastyString.c_str());
		TokenReader >> pkgid >> x >> y >> weight;
		UpdatePackage(pkgid, Point (x, y), weight, mRoboto->GetLocation());
		i+=4;
	}

	UpdateString[0] = '\0';
	for( i = 0; i < LeftOverTokens; i++)
	{
		int index = tokens.size() - LeftOverTokens + i;
		strcat(UpdateString, tokens[index].c_str());
	}

	return strlen(UpdateString);
}

int Game::HandleMessages(char * UpdateString,int nc)
{
	int Offset;
	// Parse String
	if(UpdateString[0] == '#')
		Offset = HandlePlayerUpdateMessage(UpdateString,nc);
	else
		Offset = HandlePackageUpdateMessage(UpdateString);
	return Offset;
}

void Game::Tokenize(const string& str, vector<string>& tokens, 
				    const string& delimiters)
{
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
}

void Game::EndPlay() 
{
	delete mRoboto;
}

void Game::ExecuteMove (Move & move)
{
	Point p = mRoboto->GetLocation();
	cerr << p << endl;
	cerr << move << endl;
	mIO << move << endl;
}

int Game::Valid (_asNode * parent, _asNode * node, int data, void * pointer) 
{
	Game* game = (Game*)pointer;
	if (! ((node->x >= 1) && (node->x <= game->mMapDimensions.GetX()) && (node->y >= 1) && (node->y <= game->mMapDimensions.GetY())))
		return false;

	if (parent->x - node->x != 0 && parent->y - node->y != 0) 
		return false;
	int tile = game->GetMap().Get (node->x, node->y);
	switch(tile)
	{
	case WATER:
	case WALL:
			return false;
	}
	return true;
}

int Game::Cost (_asNode * parent, _asNode *node, int data, void *pointer)
{
	Game* game = (Game*)pointer;
	switch(game->GetMap().Get (node->x, node->y))
	{
	case WATER:
	case WALL:
	case PLAIN:
	case BASE:
			return 1;
	}
	return 1;
}

CAStar* Game::GeneratePath (Point s, Point e)
{
	CAStar* star = new CAStar();


	star->Reset();
	star->SetRows (mMapDimensions.GetY() > mMapDimensions.GetX() ? 
					mMapDimensions.GetY() : mMapDimensions.GetX());
	star->udValid = Valid;
	star->udCost = Cost;
	star->udNotifyList = NULL;
	star->udNotifyChild = NULL;
	star->m_pCBData = (void *)this;
	star->m_pNCData = (void *)NULL;
	if (star->GeneratePath (e.GetX(), e.GetY(), s.GetX(), s.GetY()))
		return star;
	
	delete star;
	return (CAStar*)NULL;
}

void Game::PrintPath(_asNode* path)
{
		_asNode* node = path;
		while(node)
		{
			node = node->parent;
		}
}

Point Game::FindNearestBase ()
{
	vector<Point *> points = mMap.Find(BASE);

	cerr << "@ " << *points[0] << endl;

	return *points[0];
}	
