
#ifndef __GAME_H__
#define __GAME_H__

#include "global.h"

#include <vector>
#include <string>

#include "SparseArray.h"
#include "Package.h"
#include "Robot.h"
#include "Point.h"
#include "Robots.h"
#include "PathFinder.h"

typedef int (*GoodnessFunction)(Game *, Package *, void *);

class Game {
	private :
		Map			mMap;
		PackageList	mPackageList;
		Robots		mRobots;
		Robot*		mRoboto;
		IO			mIO;
	protected:
		Robot*		UpdateRobot (ID, MoveType);
		Robot*		UpdateRobot (ID, MoveType, Package*);
		Robot*		UpdateRobot (ID, Point);

		Package*	UpdatePackage(ID, MoveType, Robot*);
		Package*	UpdatePackage(Package*, MoveType, Robot*);
		Package* 	UpdatePackage(ID, Point, Weight, Point);

		int 		HandleMessages (char *,int);
		int 		HandlePackageUpdateMessage(char* UpdateString);
		int 		HandlePlayerUpdateMessage(char* UpdateString,int);
		void 		Tokenize(const string& str, vector<string>& tokens, 
						     const string& delimiters = " ");
		void 		InnerLoop ();	
		static int 		Valid (_asNode*,_asNode*,int,void *);
		static int 		Cost (_asNode*,_asNode*,int,void *);

	public:
		Point		mMapDimensions;
					Game(): mIO (sockbuf::sock_stream) {} ;

		bool 		BeginPlay (char * Server, int Port);
		PackageList	Query (int, GoodnessFunction = 0, void * = 0);
		CAStar*		GeneratePath (Point, Point);
		void		PrintPath(_asNode* );
		void 		ExecuteMove (Move &);
		void		EndPlay();
		Point		FindNearestBase();

		Map&		GetMap() { return mMap; };
		PackageList& GetPackageList() { return mPackageList; };
};
#endif
