#include "global.h"

#include "Move.h"
#include <string>

// SetType based on the Point p, where p is actually
// the results of p1 - p2. Since p = p1 - p2, p cannot
// be (0,0) or there would be no change. Nor can p be
// (X,Y) for that would be a diagonal move. p must be
// of the form (X,0) or (0,Y).
void Move::SetType (Point p) 
{
	if ( ((p.GetX() != 0) && (p.GetY() != 0)) ||
		 (p.GetX() == 0) && (p.GetY() == 0)) 
		SetType (NOOP);
	else if (p.GetX() == -1)
		SetType (WEST);
	else if (p.GetX() == 1)
		SetType (EAST);
	else if (p.GetY() == -1)
		SetType (SOUTH);
	else if (p.GetY() == 1)
		SetType (NORTH);

	return;
}


ostream& operator << (ostream& o, Move& m)
{
	string movetext;

	Money bid = m.GetBid();
	PackageList packages = m.GetPackageList();
	MoveType type = m.GetType();

	switch (type)
	{
		case NORTH: 
		case SOUTH: 
		case EAST : 
		case WEST : 
					movetext = "Move ";
					movetext += type;
					packages.RemoveAll();
					break;
		case DROP: 	movetext = "Drop ";
					break;
		case PICK:  movetext = "Pick ";
					break;
		case DELIVER:
					movetext = "Deliver ";
					break;
		case GATHER:
					movetext = "Gather ";
					break;
		case NOOP: 
		default  :
				   movetext = "Drop "; 
	}

	o << bid << " " << movetext << packages;
	return o;
}
