#include "PackageList.h"

Package* PackageList::Find(ID pkgid)
{
	for(unsigned int i = 0; i < mPackages.size(); i++)
		if(mPackages[i]->GetID() == pkgid)
			return mPackages[i];
	return (Package*) NULL;
}

Package* PackageList::Add (Package * package)
{
	mPackages.push_back (package);
	return (package);
}

void PackageList::Remove (Package * package)
{
	ID pkgid = package->GetID();
	for(unsigned int i = 0; i < mPackages.size(); i++)
	{
		if(mPackages[i]->GetID() == pkgid)
		{
			Package * tmp = mPackages.back();
			mPackages[i] = tmp;
			mPackages.pop_back();
			return ;
		}
	}
}

Package* PackageList::Find (Weight weight, bool fuzzy = false, bool opposite = false)
{
	Weight 	cw = INT_MAX;
	Index 	ci = 0;
	bool 	cf = false;

	for(unsigned int i = 0; i < mPackages.size(); i++)
	{
		Weight w = mPackages[i]->GetWeight();
		if(w == weight)
			return mPackages[i];
		if (fuzzy && (abs(w - weight) < cw)) {
			cw = abs (w - weight);
			ci = i; 
			cf = true;
		}
	}
	// Since cw = abs(w - weight), then cw > 0 if we found a fuzzy match
	if (cf)
		return mPackages[ci];
	return (Package*)NULL;
}

// TODO This function would be great if it could find packages that were close to a location AND
// close to a destination. For now, no go.
Package* PackageList::Find (Point location, bool fuzzy, bool opposite = false)
{
	Distance 	cd = INT_MAX;	
	Index 		ci;
	bool		cf = false;

	for(unsigned int i = 0; i < mPackages.size(); i++)
	{
		Distance d = 0;
		Point package = mPackages[i]->GetLocation();
		if (package == location) 
			return mPackages[i];
		if (fuzzy && ( (d = Point::Distance(package, location)) < cd)) {
			cd = d;
			ci = i;
			cf = true;
		}
	}
	if (cf)
		return mPackages[ci];
	return (Package*)NULL;
}


Index PackageList::Size() 
{
	return mPackages.size();
}

Weight PackageList::GetWeight() 
{
	Weight w = 0;
	for(unsigned int i = 0; i < mPackages.size(); i++)
		w += mPackages[i]->GetWeight();
	return w;
}

// Assumes all packages are going to the same destination;
Point PackageList::GetDestination() 
{
	return mPackages[0]->GetDestination();
}

ostream& operator << (ostream& o, PackageList& p)
{
	Packages packages = p.GetPackages();
	for(unsigned int i = 0; i < packages.size(); i++)
			o << packages[i]->GetID() << " ";
	return o;
}

