#include <string>

#include "global.h"
#include "Game.h"
#include "Robot.h"
#include "Package.h"
#include "Point.h"
#include "PathFinder.h"

// The point is to generate many excellent moves.
// To begin with, this function will probably be dumb, but
// constant refinement will allow it to grow to epic
// proportions of ... what is it for again?
MoveList Robot::Think (Game * game)
{
	mMoveList.RemoveAll();
	// We can always do a NOOP
	mMoveList.Add(new Move());
	Move * base = new Move();
	Point nearestbase = game->FindNearestBase();

	if (nearestbase != mLocation) 
	{
		base->SetType(GATHER);
		base->SetPackageList(new Package(-1,nearestbase,-1,nearestbase));
		base->SetGoodness(2);
		mMoveList.Add (base);
	}

	// All packages we currently own are eligible for dropping.
	// They are also available for delivering -- unless we are
	// already on their destination square.
	Packages packages = mPackageList.GetPackages();
	for (Index i = 0; i < packages.size(); i++) 
	{
		Package* package = packages[i];
		Move * move = new Move();
		move->SetType (DROP); 
		move->SetPackageList (package);
		Rate(game, move);
		mMoveList.Add (move);
		if (package->GetDestination() != this->GetLocation()) 
		{
			move = new Move();
			move->SetType (DELIVER);
			move->SetPackageList (package);
			Rate(game, move);
			mMoveList.Add (move);
		}
	}
	// All packages on the board are eligible for picking up
	// Eventually we will want to do more than just the entire list
	// of packages.
	// Refinements :
	// 	Not if they are unreachable (or too far away).
	// 	Not if we already have them.
	PackageList packagelist = game->GetPackageList();
	packages = packagelist.GetPackages();
	for (Index i = 0; i < packages.size(); i++)
	{
		if (packages[i]->GetOwner() == NULL)
		{
			Move * move = new Move();
			if(packages[i]->GetLocation() != this->GetLocation())
				move->SetType(GATHER);
			else
				move->SetType (PICK);
			move->SetPackageList (packages[i]);
			Rate(game, move); 
			mMoveList.Add (move);
		}
	}
	return mMoveList;
}

Move Robot::Decide (Game * game, MoveList movelist)
{
	Move * theMove;
	Moves moves = movelist.GetMoves();

	theMove = moves[0];

	for (Index i = 0; i < moves.size(); i++)
	{
		Move * move = moves[i];
		move->SetBid (move->GetGoodness());
		if (*move > *theMove)
		{
			theMove = move;
		}
	}


	return *theMove;
}

void Robot::Rate (Game * game, Move * move)
{
	PackageList pl = move->GetPackageList();
	Point		rl = this->GetLocation();
	Point		dest;

	switch (move->GetType())
	{
		case GATHER:
			move->SetGoodness(2);
			break;
		case DELIVER:
			//dest = pl.GetDestination();
			move->SetGoodness(3);
			break;
		case DROP:
			dest = pl.GetDestination();
			if(dest == rl)
				move->SetGoodness(15);
			else
				move->SetGoodness(1);
			break;
		case PICK:
			move->SetGoodness (4);
			break;
		default:
			move->SetGoodness (1);
			break;
	}

}

void Robot::MakeMove (Game * game) 
{

	MoveList moves = Think(game);
	mNextMove = Decide (game, moves);

	if (mNextMove.GetType() == DELIVER || mNextMove.GetType() == GATHER) 
	{
		cerr << "Transforming" << endl;
		PackageList pl = mNextMove.GetPackageList();
		Point dest = pl.GetDestination();

		CAStar* path = game->GeneratePath (mLocation,dest);
		cerr << "Attempting path find from " << mLocation << " to " << dest << endl;
		if (path) 
		{
			cerr << "Got Path" << endl;
			_asNode * node = path->GetBestNode();
			// For some reason, GeneratePath always gives us the ending node. So game->GeneratePath
			// secretly reverses start and finish for us, to make things easier. In addition, we can skip
			// over the first node, which is, predictably, the current.
			game->PrintPath(node);
			if (node)
				node = node->parent;
			if (node != NULL) 
			{
				Point p = Point (node->x, node->y);

				mNextMove.SetType(p - mLocation);
				mNextMove.SetBid(1);
			}
		}
	}
	game->ExecuteMove (mNextMove);
}

void Robot::Update (MoveType direction) 
{

	switch (direction) 
	{
		case NORTH : mLocation.SetY(mLocation.GetY()+1); break;
		case SOUTH : mLocation.SetY(mLocation.GetY()-1); break;
		case WEST  : mLocation.SetX(mLocation.GetX()-1); break;
		case EAST  : mLocation.SetX(mLocation.GetX()+1); break;
		default :
			break;
	}
}

Package* Robot::Update (MoveType action, Package* package)
{
	switch (action) 
	{
		case PICK : mPackageList.Add (package); break;
		case DROP : mPackageList.Remove (package); break;
		default :
			break;
	}
	return package;
}

ostream& operator << (ostream& o, Robot &r)
{
	o << r.GetID();


	return o;
}

ostream& operator << (ostream& o, Robot *r)
{
	if (r)
		o << *r;
	else
		o << "Robot nil";

	return o;
}

